// Copyright 2016-2018 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "tokenpool-gnu-make.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define DEBUG(S) printf(S); // TODO varargs

#if 0
#include "line_printer.h" // ninja.cc
#define PRINT(S) { LinePrinter printer; printer.PrintOnNewLine(S); }
#else
#define PRINT(S) printf(S);
#endif

using namespace std;

// TokenPool implementation for GNU make jobserver - common bits
// every instance owns an implicit token -> available_ == 1
GNUmakeTokenPool::GNUmakeTokenPool() : available_(1), used_(0) {
}

GNUmakeTokenPool::~GNUmakeTokenPool() {
}

bool GNUmakeTokenPool::SetupClient(bool ignore,
                                   bool verbose,
                                   double& max_load_average) {
  const char* value = GetEnv("MAKEFLAGS");
  if (!value) {
    DEBUG("GNUmakeTokenPool::SetupClient: init failed. MAKEFLAGS is empty\n");
    return false;
  }

  // GNU make <= 4.1
  const char* jobserver = strstr(value, "--jobserver-fds=");
  if (!jobserver)
    // GNU make => 4.2
    jobserver = strstr(value, "--jobserver-auth=");
  if (jobserver) {
    //LinePrinter printer;

    DEBUG("GNUmakeTokenPool::SetupClient: found --jobserver-fds or --jobserver-auth in MAKEFLAGS\n");

    if (ignore) {
      PRINT("ninja: warning: -jN forced on command line; ignoring GNU make jobserver.\n");
    } else {
      if (ParseAuth(jobserver)) {
        const char* l_arg = strstr(value, " -l");
        int load_limit = -1;

        if (verbose) {
          PRINT("ninja: using GNU make jobserver.\n");
        }

        // translate GNU make -lN to ninja -lN
        if (l_arg &&
            (sscanf(l_arg + 3, "%d ", &load_limit) == 1) &&
            (load_limit > 0)) {
          max_load_average = load_limit;
        }

        return true;
      }
    }
  }

  return false;
}

bool GNUmakeTokenPool::SetupMaster(bool verbose,
                                   int parallelism,
                                   double max_load_average) {
  // no need to set up token pool for serial builds
  if (parallelism == 1)
    return false;

  string auth;
  if (CreatePool(parallelism, &auth)) {
    string value = "--jobserver-auth=" + auth;
    {
      char buffer[32];
      snprintf(buffer, sizeof(buffer), "%i", parallelism);
      value += " -j";
      value += buffer;
    }
    if (max_load_average > 0.0f) {
      char buffer[32];
      snprintf(buffer, sizeof(buffer), "%g", max_load_average);
      value += " -l";
      value += buffer;
    }

    if (SetEnv("MAKEFLAGS", value.c_str())) {
      if (verbose) {
        //LinePrinter printer;
        PRINT("ninja: simulating GNU make jobserver.\n");
      }
      return true;
    }
  }

  return false;
}

bool GNUmakeTokenPool::Acquire() {
  printf("GNUmakeTokenPool::Acquire: available_ = %i\n", available_);
  if (available_ > 0)
    return true;

  int token;
  if ((token = AcquireToken()) >= 0) {
    // token acquired
    printf("GNUmakeTokenPool::Acquire: token acquired: %i\n", token);
    available_++;
    return true;
  }

  printf("GNUmakeTokenPool::Acquire: no token available\n");
  // no token available
  return false;
}

void GNUmakeTokenPool::Reserve() {
  available_--;
  used_++;
}

void GNUmakeTokenPool::Return() {
  if (ReleaseToken())
    available_--;
}

void GNUmakeTokenPool::Release() {
  available_++;
  used_--;
  if (available_ > 1)
    Return();
}

void GNUmakeTokenPool::Clear() {
  while (used_ > 0)
    Release();
  while (available_ > 1)
    Return();
}
